/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.registry_overlay.junit;

import java.math.BigInteger;
import java.util.Properties;

import org.apache.mina.util.AvailablePortFinder;
import org.junit.Rule;
import org.junit.rules.ExternalResource;
import org.ow2.petals.registry_overlay.api.Constants;
import org.ow2.petals.registry_overlay.configuration.generated.Credentials;
import org.ow2.petals.registry_overlay.configuration.generated.Member;
import org.ow2.petals.registry_overlay.configuration.generated.Members;
import org.ow2.petals.registry_overlay.configuration.generated.PetalsRegistryOverlay;
import org.ow2.petals.registry_overlay.core.MemberConstants;
import org.ow2.petals.registry_overlay.core.RegistryOverlayNode;
import org.ow2.petals.registry_overlay.core.exception.ConfigurationException;

/**
 * The {@link PetalsRegistryOverlayNode} {@link Rule} allows creation of a Petals Registry Overlay node. It is
 * guaranteed to be deleted when the test method finishes (whether it passes or fails):
 * 
 * <pre>
 * public class MyTestSuite {
 *  &#064;Rule
 *  public PetalsRegistryOverlayNode node = new PetalsRegistryOverlayNode(...);
 * 
 *  &#064;Test
 *  public void test...() throws IOException {
 *      ...
 *  }
 * }
 * </pre>
 */
public class PetalsRegistryOverlayNode extends ExternalResource {

    private static final String MEMBER_ID_STANDALON_CLUSTER = "member-of-a-standalone-cluster";

    /**
     * The configuration of the Petals Registry Node
     */
    private final PetalsRegistryOverlay petalsRegistryOverlayNodeCfg;

    /**
     * Properties of the Petals Registry Node
     */
    private final Properties petalsRegistryOverlayNodeProps;

    /**
     * The member identifier of the Petals Registry node
     */
    private final String memberId;

    /**
     * The Petals Registry node
     */
    private RegistryOverlayNode petalsRegistryNode = null;

    /**
     * Creates a temporary Petals Registry Overlay node with the given configuration.
     * 
     * @param port
     *            The listening port. <code>0</code>: the port will be computed on startup.
     * @param groupName
     *            The group part of the credentials
     * @param password
     *            The password part of the credentials
     * @param jmxPort
     *            The listening port of the JMX connector. <code>0</code> disables the JMX support.
     */
    public PetalsRegistryOverlayNode(final int port, final String groupName, final String password, final int jmxPort) {
        
        this.memberId = MEMBER_ID_STANDALON_CLUSTER;
        this.petalsRegistryOverlayNodeCfg = new PetalsRegistryOverlay();
        
        final Credentials credentials = new Credentials();
        credentials.setGroup(groupName);
        credentials.setPassword(password);
        this.petalsRegistryOverlayNodeCfg.setCredentials(credentials);
        
        final Member member = new Member();
        member.setId(this.memberId);
        member.setPort(BigInteger.valueOf(port));
        member.setValue("localhost");
        final Members members = new Members();
        members.getMember().add(member);
        this.petalsRegistryOverlayNodeCfg.setMembers(members);
        
        this.petalsRegistryOverlayNodeProps = new Properties();
        this.petalsRegistryOverlayNodeProps.setProperty(MemberConstants.MEMBER_PROPERTY_NAME_MEMBER_ID, this.memberId);

        if (jmxPort == 0) {
            this.petalsRegistryOverlayNodeProps.setProperty(MemberConstants.MEMBER_PROPERTY_NAME_DISABLE_JMX_CONNECTOR,
                    Boolean.FALSE.toString());
        } else {
            this.petalsRegistryOverlayNodeProps.setProperty(MemberConstants.MEMBER_PROPERTY_NAME_JMX_PORT,
                    String.valueOf(jmxPort));
        }
    }

    /**
     * Creates a temporary Petals Registry Overlay node with the given cluster configuration. If the listening port of
     * the node is set to <code>0</code>, a free port will be computed on startup.
     * 
     * @param petalsRegistryOverlay
     *            The cluster configuration of the Petals Registry Overlay
     * @param memberId
     *            The member identifier of the Petals Registry Overlay node to create
     * @throws ConfigurationException
     *             An error about the Petals Registry Overlay configuration occurs
     */
    public PetalsRegistryOverlayNode(final PetalsRegistryOverlay petalsRegistryOverlay, final String memberId)
            throws ConfigurationException {

        this.petalsRegistryOverlayNodeCfg = petalsRegistryOverlay;
        this.memberId = memberId;

        this.petalsRegistryOverlayNodeProps = new Properties();
        this.petalsRegistryOverlayNodeProps.setProperty(MemberConstants.MEMBER_PROPERTY_NAME_MEMBER_ID, this.memberId);
    }

    /**
     * Creates a temporary Petals Registry Overlay node with a default configuration:
     * <ul>
     * <li>protected by the default credentials: {@value Constants#DEFAULT_CREDENTIALS_GROUP}/
     * {@value Constants#DEFAULT_CREDENTIALS_PASSWORD}</li>
     * <li>JMX is disabled.</li>
     * 
     * @param port
     *            The listening port. <code>0</code>: the port will be computed on startup.
     * 
     */
    public PetalsRegistryOverlayNode(final int port) {
        this(port, MemberConstants.DEFAULT_CREDENTIALS_GROUP, MemberConstants.DEFAULT_CREDENTIALS_PASSWORD, 0);
    }

    /**
     * Creates a temporary Petals Registry Overlay node with a default configuration:
     * <ul>
     * <li>listening on a free port starting at {@value Constants#DEFAULT_PORT},</li>
     * <li>protected by the default credentials: {@value Constants#DEFAULT_CREDENTIALS_GROUP}/
     * {@value Constants#DEFAULT_CREDENTIALS_PASSWORD}</li>
     * <li>JMX is disabled.</li>
     */
    public PetalsRegistryOverlayNode() {
        this(0, MemberConstants.DEFAULT_CREDENTIALS_GROUP, MemberConstants.DEFAULT_CREDENTIALS_PASSWORD, 0);
    }

    /**
     * Creates a temporary Petals Registry Overlay node with the given credentials:
     * <ul>
     * <li>listening on a free port starting at {@value Constants#DEFAULT_PORT},</li>
     * <li>JMX is disabled.</li>
     */
    public PetalsRegistryOverlayNode(final String groupName, final String password) {
        this(0, groupName, password, 0);
    }

    @Override
    protected void before() throws Throwable {

        this.startPetalsRegistryNode();
    }

    @Override
    protected void after() {
        try {
            this.stop();
        } catch (final InterruptedException e) {
            e.printStackTrace();
        }
    }

    /**
     * Initialize & start the Petals Registry node configured
     */
    private void startPetalsRegistryNode() {

        for (final Member member : this.petalsRegistryOverlayNodeCfg.getMembers().getMember()) {
            if (this.memberId.equals(member.getId())) {
                if (0 == member.getPort().intValue()) {
                    member.setPort(
                            BigInteger.valueOf(AvailablePortFinder.getNextAvailable(MemberConstants.DEFAULT_PORT)));
                }
                break;
            }
        }

        try {
            this.petalsRegistryNode = new RegistryOverlayNode(this.petalsRegistryOverlayNodeCfg,
                    this.petalsRegistryOverlayNodeProps);
        } catch (final ConfigurationException e) {
            throw new RuntimeException("A bug occurs in the configuration of the Petals Registry node", e);
        }

        this.start();
    }

    /**
     * Start the Petals Registry node
     * 
     * @throws InterruptedException
     */
    public void start() {

        if (this.petalsRegistryNode != null && !this.isRunning()) {
            this.petalsRegistryNode.start();
        }
    }

    /**
     * Stop the Petals Registry node
     * 
     * @throws InterruptedException
     */
    public void stop() throws InterruptedException {

        // Wait end of HZL propagation processing
        Thread.sleep(1000);

        if (this.petalsRegistryNode != null && this.isRunning()) {
            this.petalsRegistryNode.stop();
        }
    }

    /**
     * @return <code>true</code> if the Petals Registry Node is running
     */
    public boolean isRunning() {
        return this.petalsRegistryNode.isRunning();
    }

    /**
     * @return The hostname on which the Petals Registry node is running, or <code>null</code> if the current member is
     *         not found in the cluster definition.
     */
    public String getHost() {
        for (final Member member : this.petalsRegistryOverlayNodeCfg.getMembers().getMember()) {
            if (this.memberId.equals(member.getId())) {
                return member.getValue();
            }
        }
        return null;
    }

    /**
     * @return The listening port of the Petals Registry node, or <code>0</code> if the current member is not found in
     *         the cluster definition.
     */
    public int getPort() {
        for (final Member member : this.petalsRegistryOverlayNodeCfg.getMembers().getMember()) {
            if (this.memberId.equals(member.getId())) {
                return member.getPort().intValue();
            }
        }
        return 0;
    }

    /**
     * @return The group part of the credentials of the Petals Registry node
     */
    public String getGroupName() {
        return this.petalsRegistryOverlayNodeCfg.getCredentials().getGroup();
    }

    /**
     * @return The password part of the credentials of the Petals Registry node
     */
    public String getPassword() {
        return this.petalsRegistryOverlayNodeCfg.getCredentials().getPassword();
    }
}
